from typing import Iterable, Iterator, Callable
from hashlib import sha256
from functools import reduce


def dhash256(x: int) -> int:
    return sha256(sha256(x).digest()).digest()


def trunks(l: Iterable, n: int) -> Iterator:
    return zip(*[iter(l)] * n)


def concat(a: bytes, b: bytes) -> bytes:
    return bytes(a) + bytes(b)


def compose(*fs: Iterable[Callable]):
    def _(*args, **kwargs):
        return reduce(
            lambda x, y: y(x),
            fs,
            kwargs or args
        )
    return _


def to_sha256int(m: str):
    return int.from_bytes(sha256(m.encode()).digest(), 'big')


class EnumDict(dict):
    def __getattr__(self, k):
        '''
        Support Lazy
        '''
        value = self.get(k)
        if callable(value):
            value = value()
        if type(value) is dict:
            return EnumDict(value)
        else:
            return value

    def __setattr__(self, k, v):
        self.update({k: v})
        return self

    def __contains__(self, v):
        return v in self.values()



import base64
import base58
from klefki.types.algebra.isomorphism import bijection
from klefki.types.algebra.concrete import (
    FiniteFieldCyclicSecp256k1 as CF,
)


b58encode = bijection(base58.b58decode)(base58.b58encode)
b64encode = bijection(base64.b64decode)(base64.b64encode)


def int_to_byte(key: int) -> bytes:
    return key.to_bytes(32, byteorder='big')


@bijection(int_to_byte)
def byte_to_int(byte: bytes) -> CF:
    return int(byte.hex(), 16)



from .groups import EllipticCurveGroup, CyclicGroup, JacobianGroup
from .fields import FiniteField
import klefki.const as const

__all__ = [
    'FiniteFieldSecp256k1',
    'FiniteFieldCyclicSecp256k1',
    'EllipticCurveGroupSecp256k1',
    'JacobianGroupSecp256k1',
    'EllipticCurveCyclicSubgroupSecp256k1'
]


class FiniteFieldSecp256k1(FiniteField):
    __slots__ = ()
    P = const.P


class FiniteFieldCyclicSecp256k1(FiniteField):
    __slots__ = ()
    P = const.N


class EllipticCurveGroupSecp256k1(EllipticCurveGroup):
    __slots__ = ()
    A = const.A
    B = const.B


class JacobianGroupSecp256k1(JacobianGroup):
    __slots__ = ()
    A = const.A
    B = const.B


class EllipticCurveCyclicSubgroupSecp256k1(CyclicGroup):
    __slots__ = ()
    G = EllipticCurveGroupSecp256k1(
        (
            FiniteFieldSecp256k1(const.Gx),
            FiniteFieldSecp256k1(const.Gy)
        )
    )
    N = const.N



from sha3 import keccak_256
from .public import encode_pubkey
from klefki.types.algebra.concrete import (
    EllipticCurveGroupSecp256k1 as ECG,
)


def gen_address(pub: ECG) -> str:
    return gen_address_from_str(pub)


def gen_address_from_str(pub: str) -> str:
    return keccak_256(pub).hexdigest



def command(fn, name=None):

    if not hasattr(command, 'registered'):
        command.registered = {}
    fname = name or fn.__name__
    command.registered[fname] = fn

    return fn


def not_found():
    print('cmd not found')
    return


def router(cmd: str):
    'fn a b c=1 d=2'
    cmds = cmd.split(' ')
    fn = command.registered.get(cmds[0], None)
    if not fn:
        return not_found()
    args = [c for c in cmds if '=' not in cmds[1:]]
    kwargs = dict([c.split('=') for c in cmds if '=' in cmds[1:]])
    return fn(*args, **kwargs)



from hashlib import sha256
import hashlib
import base58
from klefki.crypto.ecdsa import (
    pubkey
)
from klefki.types.algebra.concrete import (
    EllipticCurveGroupSecp256k1 as ECG,
    EllipticCurveCyclicSubgroupSecp256k1 as CG,
    FiniteFieldSecp256k1 as F
)

__all__ = [
    'gen_address'
]


G = CG.G
A = ECG.A
B = ECG.B
P = F.P


def gen_address(pub: ECG) -> str:
    x, y = pub.value[0].value, pub.value[1].value
    if y % 2 == 0:
        prefix = bytes([0x02])
    else:
        prefix = bytes([0x03])
    networkid = bytes([0x00])
    ripemd160 = hashlib.new('ripemd160')
    ripemd160.update(
        sha256(
            prefix + x.to_bytes(32, byteorder='big')
        ).digest()
    )
    hashed = ripemd160.digest()
    assert len(hashed) == 20
    with_network = networkid + hashed
    auth = sha256(sha256(with_network).digest()).digest()[:4]
    res = with_network + auth
    assert len(res) == 25
    return base58.b58encode(res)


def gen_address_from_key(key) -> str:
    return gen_address(pubkey(key))
